unit DirectoryCompare;

interface

uses
  SysUtils;

type
  EDirectoryNotExisting = class(Exception);
  EDirectoriesAlreadyDifferent = class(Exception);

function CompareDirectories(A, B: string): boolean;

implementation

uses
  FileCompare;

resourcestring
  ERR_DIR_NOT_EXIST = 'Directory "%s" does not exist.';
  ERR_ALREADY_DIFFERENT = 'The two directories have not the same number of files and directories.';

procedure CheckDir(var A: string);
begin
  A := IncludeTrailingPathDelimiter(A);
  if not DirectoryExists(A) then
  begin
    raise EDirectoryNotExisting.CreateFmt(ERR_DIR_NOT_EXIST, [A]);
  end;
end;

function FilesInAButNotInBExisting(A, B: string): boolean;
var
  SR: TSearchRec;
begin
  CheckDir(A);
  CheckDir(B);

  result := false;

  if FindFirst(A+'*', faAnyFile, SR) = 0 then
  begin
    repeat
      if (SR.Name <> '.') and (SR.Name <> '..') then
      begin
        if DirectoryExists(A+SR.Name) and DirectoryExists(B+SR.Name) then
        begin
          if FilesInAButNotInBExisting(A+SR.Name, B+SR.Name) then
          begin
            result := true;
            exit;
          end;
        end
        else if not FileExists(B+SR.Name) then
        begin
          result := true;
          exit;
        end;
      end;
    until FindNext(SR) <> 0;
    FindClose(SR);
  end;
end;

function DifferentFilesInDirectories(A, B: string): boolean;
var
  SR: TSearchRec;
begin
  CheckDir(A);
  CheckDir(B);

  result := false;

  if FindFirst(A+'*', faAnyFile, SR) = 0 then
  begin
    repeat
      if (SR.Name <> '.') and (SR.Name <> '..') then
      begin
        if DirectoryExists(A+SR.Name) then
        begin
          if not DirectoryExists(B+SR.Name) then
          begin
            raise EDirectoriesAlreadyDifferent.Create(ERR_ALREADY_DIFFERENT);
          end;

          if DifferentFilesInDirectories(A+SR.Name, B+SR.Name) then
          begin
            result := true;
            exit;
          end;
        end
        else
        begin
          if not FileExists(B+SR.Name) then
          begin
            raise EDirectoriesAlreadyDifferent.Create(ERR_ALREADY_DIFFERENT);
          end;

          if not CompareFilePhysical(A+SR.Name, B+SR.Name) then
          begin
            result := true;
            exit;
          end;
        end;
      end;
    until FindNext(SR) <> 0;
    FindClose(SR);
  end;
end;

function CompareDirectories(A, B: string): boolean;
begin
  result := true;

  CheckDir(A);
  CheckDir(B);

  // Gibt es Dateien/Ordner in Verzeichnis A, die es in Verzeichnis B nicht gibt?
  // -> Verzeichnisse unterschiedlich
  if FilesInAButNotInBExisting(A, B) then
  begin
    result := false;
    exit;
  end;

  // Gibt es Dateien/Ordner in Verzeichnis B, die es in Verzeichnis A nicht gibt?
  // -> Verzeichnisse unterschiedlich
  if FilesInAButNotInBExisting(B, A) then
  begin
    result := false;
    exit;
  end;

  // Aktueller Stand: Beide Verzeichnisse haben die gleichen Dateien und Ordner
  // Sind die Dateien auch vom Dateiinhalt her gleich?
  if DifferentFilesInDirectories(A, B) then
  begin
    result := false;
    exit;
  end;
end;

end.
