program fcmove;

(*
 * File Compare'n'move Utility FC-MOVE 1.2
 * (C) Copyright 2006 - 2009 ViaThinkSoft.
 *
 * Zukunft:
 * - Directory Vergleich: Wird es evtl Probleme bei NTFS-Hardlinks geben (Vista-Rekursion)?
 * - Mglichkeit, Thumbs.db zu missachten etc.? (-> ggf. neues Tool fr das rekursiven Lschen von *.~* oder Thumbs.db)
 * - Kann man prfen, ob der Source-Name berhaupt ein gltiger Dateiname ist? (Das selbe fr Dest)
 *   -> In diesem Falle auch sofort die Help-Seite zeigen.
 *)

{$APPTYPE CONSOLE}

{$WARNINGS OFF}

uses
  SysUtils,
  Windows,
  FileCompare,
  DirectoryCompare;

type
  TRecursiveMode = (rmNone, rmEasy, rmMerge);

  ENotADirectory = class(Exception);

const
  NewNameStructure = '%s (%d)%s';
  f_all = '*';

resourcestring
  lng_header          = 'File Compare''n''move Utility FC-MOVE 1.2'+#13#10+
                        '(C) Copyright 2006 - 2009 ViaThinkSoft [www.viathinksoft.com].';
  lng_identical_error = 'Fatal error: The directory of the source file and the destination directory'+#13#10+
                        'may not be the same!';
  lng_not_a_directory = 'Error: The directory "%s" does not exists.';
  lng_nothing_found   = 'No files with such a search criteria.';
  lng_force_error     = 'Fatal error: Could not create the target directory recursively.';
  lng_help            = 'This utility moves files and checks if a destination file has the same name.'+#13#10+
                        'If this file has additionally the same content, the source file will be'+#13#10+
                        'deleted because the movement is not neccessary. If the files are not equal,'+#13#10+
                        'they will be renamed, so that no data gets lost.'+#13#10#13#10+
                        'Usage:'+#13#10#13#10+'FCMOVE [/De | /Dm] source destination'+#13#10#13#10+
                        '/De'+#9#9+'Also moves matching directories. They will be compared with'+#13#10+
                              #9#9+'the "easy mode". If all subdirectories and files are equal,'+#13#10+
                              #9#9+'the directory will be renamed and moved.'+#13#10+
                              #9#9+'The directories will not be merged.'+#13#10+
                        '/Dm'+#9#9+'Also moves directories. They will be merged.'+#13#10+
                              #9#9+'In this recursion, all files inside the matching folders'+#13#10+
                              #9#9+'will be compared and renamed if they are not equal.'+#13#10+
                        'source'+#9#9+'The source file(s) which should be moved to destination.'+#13#10+
                                 #9#9+'Wildcards are allowed.'+#13#10+
                        'destination'+#9+'The target directory where the file(s) should be moved to.';
  lng_identical       = '[Deleted]';
  lng_renamed         = '[Renamed]';
  lng_moved           = '[Moved]';
  lng_move_error      = '[MOVE ERROR]';
  lng_no_dirs         = '[IGNORED]';
  lng_del_error       = '[DELETE ERROR]';
  lng_errors          = 'Warning: Some files could not be moved to its destination!';
  lng_mode            = 'Mode:'+#9;
  lng_mode_easy       = 'Directory support - Easy recursive mode';
  lng_mode_merge      = 'Directory support - Merge mode';
  lng_mode_normal     = 'Standard mode (no directories)';

function HasDelimiterAtEnd(FileName: string): boolean;
begin
  result := IsPathDelimiter(FileName, Length(FileName));
end;

function AreFilesOrDirectoriesIdentical(A, B: string): boolean;
begin
  // Die Existenz macht in dieser Funktion keinen Unterschied,
  // es geht nur um die Struktur...
  // if DirectoryExists(A) or DirectoryExists(B) then

  if HasDelimiterAtEnd(A) or HasDelimiterAtEnd(B) then
  begin
    A := IncludeTrailingPathDelimiter(A);
    B := IncludeTrailingPathDelimiter(B);
  end;

  result := SameFileName(ExpandFileName(A), ExpandFileName(B));
end;

function ExtractBaseName(FileName: string): string;
begin
  result := ExtractFileName(FileName);
  result := ChangeFileExt(result, '');           
end;

function IsDirectoryEmpty(Dir: string): boolean;
var
  SR: TSearchRec;
begin
  result := true;

  if not DirectoryExists(Dir) then
  begin
    raise ENotADirectory.CreateFmt(lng_not_a_directory, [Dir]);
  end;

  Dir := IncludeTrailingPathDelimiter(Dir);

  if FindFirst(Dir + f_all, faAnyFile, SR) = 0 then
  begin
    repeat
      if (SR.Name <> '.') and (SR.Name <> '..') then
      begin
        result := false;
        exit;
      end;
    until FindNext(SR) <> 0;
    SysUtils.FindClose(SR);
  end;
end;

function RemoveDirectoryRecursive(Dir: string): boolean;
var
  SR: TSearchRec;
begin
  if not DirectoryExists(Dir) then
  begin
    result := false;
    exit;
  end;

  Dir := IncludeTrailingPathDelimiter(Dir);

  result := true;

  if FindFirst(Dir + f_all, faAnyFile, SR) = 0 then
  begin
    repeat
      if (SR.Name <> '.') and (SR.Name <> '..') then
      begin
        if DirectoryExists(Dir + SR.Name) then
        begin
          RemoveDirectoryRecursive(Dir + SR.Name); // Rekursiver Aufruf
        end
        else
        begin
          // Hier prfen wir nicht auf Erfolg. Die Fehlermeldung wird nmlich
          // beim RemoveDirectory() des erwarteten ausgegeben.
          DeleteFile(PChar(Dir + SR.Name));
        end;
      end;
    until FindNext(SR) <> 0;
    SysUtils.FindClose(SR);
  end;

  if IsDirectoryEmpty(Dir) then // Sicher ist sicher...
  begin
    if not RemoveDirectory(PChar(Dir)) then
    begin
      result := false;
    end;
  end
  else
  begin
    result := false;
  end;
end;

procedure _Recursive(ASource, ADestination, AOutPrefix: string;
  ARecursiveMode: TRecursiveMode; var RErrors, RFound: boolean);
var
  SR: TSearchRec;
  SourceFile, TargetFile: string;
  TempName, TempName2: string;
  i: integer;
begin
  ADestination := IncludeTrailingPathDelimiter(ADestination);

  if FindFirst(ASource, faAnyFile, SR) = 0 then
  begin
    repeat
      if (SR.Name <> '.') and (SR.Name <> '..') then
      begin
        SourceFile := ExtractFilePath(ASource) + SR.Name;
        TargetFile := ADestination + SR.Name;

        if DirectoryExists(PChar(SourceFile)) then
        begin
          // *** ORDNER ***

          if ARecursiveMode = rmEasy then
          begin
            Write(AOutPrefix + SR.Name + PathDelim);
            Write(' ');

            // Existiert bereits ein Verzeichnis mit diesem Namen?
            // (Es folgt Analoger Code zu der Dateiroutine)
            if DirectoryExists(TargetFile) then
            begin
              if CompareDirectories(SourceFile, TargetFile) then
              begin
                // Gleicher Ordner mit gleichem Namen existiert bereits!

                if not RemoveDirectoryRecursive(SourceFile) then
                begin
                  WriteLn(lng_del_error);
                  RErrors := true;
                end
                else
                begin
                  WriteLn(lng_identical);
                end;
              end
              else
              begin
                // Ordner mit anderem Namen, aber anderem Inhalt existiert
                // bereits! Suche einen neuen Namen
                i := 1;
                repeat
                  Inc(i);
                  TempName := ADestination + Format(NewNameStructure,
                              [ExtractBaseName(TargetFile), i,
                               ExtractFileExt(TargetFile)]);
                until not DirectoryExists(TempName);

                if not MoveFile(PChar(SourceFile), PChar(TempName)) then
                begin
                  WriteLn(lng_move_error);
                  RErrors := true;
                end
                else
                begin
                  WriteLn(lng_renamed);
                end;
              end;
            end
            else
            begin
              // Ordner mit gleichem Namen existiert nicht, verschiebe
              if not MoveFile(PChar(SourceFile), PChar(TargetFile)) then
              begin
                WriteLn(lng_move_error);
                RErrors := true;
              end
              else
              begin
                WriteLn(lng_moved);
              end;
            end;

            RFound := true;
          end
          else if ARecursiveMode = rmMerge then
          begin
            TempName  := SourceFile + PathDelim + f_all;
            TempName2 := ADestination + SR.Name + PathDelim;

            if IsDirectoryEmpty(SourceFile) then
            begin
              // Ein Leeres Verzeichnis wird verschoben

              Write(AOutPrefix + SR.Name + PathDelim);
              Write(' ');

              if not MoveFile(PChar(SourceFile), PChar(TempName2)) then
              begin
                WriteLn(lng_move_error);
                RErrors := true;
              end
              else
              begin
                WriteLn(lng_moved);
              end;
            end
            else
            begin
              // Ein Verzeichnis wird gemerg'ed

              _Recursive(TempName,
                         TempName2,
                         AOutPrefix + SR.Name + PathDelim,
                         ARecursiveMode, RErrors, RFound);

              // Ist das Verzeichnis nun leer, lschen wir es.

              // Hier werden keine Fehlermeldung ausgegeben. Denn eine
              // Fehlermeldung wurde bereits bei DeleteFile() ausgegeben.

              if IsDirectoryEmpty(SourceFile) then
              begin
                RemoveDirectory(PChar(SourceFile));
              end;
            end;
          end
          else // if ARecursiveMode = rmNone
          begin
            Write(AOutPrefix + SR.Name + PathDelim);
            Write(' ');
            WriteLn(lng_no_dirs);
          end;
        end
        else
        begin
          // *** DATEIEN ***

          Write(AOutPrefix + SR.Name);
          Write(' ');

          // Existiert bereits eine Datei mit diesem Namen?
          if FileExists(TargetFile) then
          begin
            if CompareFilePhysical(SourceFile, TargetFile) then
            begin
              // Gleiche Datei mit gleichem Namen existiert bereits!
              if not DeleteFile(PChar(SourceFile)) then
              begin
                WriteLn(lng_del_error);
                RErrors := true;
              end
              else
              begin
                WriteLn(lng_identical);
              end;
            end
            else
            begin
              // Datei mit anderem Namen, aber anderem Inhalt existiert bereits!
              // Suche einen neuen Namen
              i := 1;
              repeat
                Inc(i);
                TempName := ADestination + Format(NewNameStructure,
                            [ExtractBaseName(TargetFile), i,
                             ExtractFileExt(TargetFile)]);
              until not FileExists(TempName);

              if not MoveFile(PChar(SourceFile), PChar(TempName)) then
              begin
                WriteLn(lng_move_error);
                RErrors := true;
              end
              else
              begin
                WriteLn(lng_renamed);
              end;
            end;
          end
          else
          begin
            // Datei mit gleichem Namen existiert nicht, verschiebe
            if not MoveFile(PChar(SourceFile), PChar(TargetFile)) then
            begin
              WriteLn(lng_move_error);
              RErrors := true;
            end
            else
            begin
              WriteLn(lng_moved);
            end;
          end;

          RFound := true;
        end;
      end;
    until FindNext(SR) <> 0;
    SysUtils.FindClose(SR);
  end;
end;

var
  Dest, Source: string;
  recmode: TRecursiveMode;
  found, errors: boolean;

begin
  writeln(lng_header);
  writeln('');

  if (ParamCount < 2) or (ParamCount > 3) then
  begin
    WriteLn(lng_help);
    WriteLn('');
    exit;
  end;

  if ParamStr(1) = '/De' then
    recmode := rmEasy
  else if ParamStr(1) = '/Dm' then
    recmode := rmMerge
  else
    recmode := rmNone;

  if recmode = rmNone then
  begin
    Source := ParamStr(1);
    Dest   := ParamStr(2);
  end
  else
  begin
    Source := ParamStr(2);
    Dest   := ParamStr(3);
  end;

  Write(lng_mode);
  if recMode = rmEasy then
    WriteLn(lng_mode_easy)
  else if recMode = rmMerge then
    WriteLn(lng_mode_merge)
  else
    WriteLn(lng_mode_normal);
  WriteLn('');

  if not ForceDirectories(Dest) then
  begin
    WriteLn(lng_force_error);
    WriteLn('');
    exit;
  end;

  if AreFilesOrDirectoriesIdentical(ExtractFilePath(Source), Dest) then
  begin
    WriteLn(lng_identical_error);
    WriteLn('');
    exit;
  end;

  found := false;
  errors := false;

  _Recursive(Source, Dest, '', recmode, errors, found);

  if not found then
  begin
    WriteLn(lng_nothing_found);
  end;

  WriteLn('');

  if errors then
  begin
    WriteLn(lng_errors);
    WriteLn('');
  end;
end.
